import itebd_functions
import pylab as pl
import numpy as np
import scipy as sp
import scipy.sparse.linalg.eigen.arpack as arp
from scipy import integrate
from scipy.linalg import expm
from svd_robust import svd_gesvd
from scipy.sparse.linalg import eigs
np.set_printoptions(linewidth=2000, precision=3,threshold=4000)

######## Define the simulation parameter ######################
d = 2  # physical dimension
chi_max = 50 # maximal bond dimension, typically between a few tens and some hundreds
N = 1000 # number of iTEBD iterations to be performed
delta = 0.01 # (imaginary) time step to be used
tol= 10.**(-15) # tolerance for truncation in the bond dimension
L=2 # unit cell length for the simulation, 2 at least

g_list = np.arange(0,2.6,0.1) # coupling parameter of the Ising model, i.e., transverse field
sx = np.array([[0.,1.],[1.,0.]])
sy = np.array([[0.,-1J],[1J,0.]])
sz = np.array([[1.,0.],[0.,-1.]])
s0 = np.array([[1.,0.],[0.,1.]])

S_list = []
m_list = []
E0_exact_list = []
E_list = []
dE_list = []
ES_list = []
xi_list = []

for g in g_list:
    ################# Get the exact energies, see Lieb-Mattis or Pfeuty ######################
    f = lambda k,g : -2*np.sqrt(1+g**2-2*g*np.cos(k))/np.pi/2.
    E0_exact = integrate.quad(f, 0, np.pi, args=(g,))[0]
    E0_exact_list.append(E0_exact)
    
    ############### Define the initial state: FM state ############
    # one could create a service subroutine out of this
    B=[]; s=[];
    for i in range(L):
        B.append(np.zeros([d,1,1])); 
        B[-1][0,0,0]=1
        s.append(np.ones([1]))        

    ########### Define Ising Hamiltonian and get U ################
    H=[]; U=[];
    Htmp=-np.kron(sz,sz) + g*np.kron(sx,np.eye(d))
    for i in range(L):
        H.append(Htmp)
        U.append(np.reshape(expm(-delta*Htmp),(d,d,d,d)))
    
    ################# Do the iTEBD run ############################
    for step in range(1, N):
        itebd_functions.update(B,s,U,chi_max,tol)
    ################# Measure stuff and store results ############
    S=np.mean(itebd_functions.entanglement_entropy(s,tol))
    S_list.append(S)
    ES=itebd_functions.entanglement_spectrum(s,tol)
    ES_list.append(ES)
    
    m=np.mean(np.abs(itebd_functions.site_expectation_value(B,s,sz)))
    m_list.append(m)
    
    E=np.mean(itebd_functions.bond_expectation_value(B,s,np.reshape(H[0],(d,d,d,d))))
    E_list.append(E)
    
    dE_list.append(np.abs((E-E0_exact)/E0_exact))
    
    xi=itebd_functions.correlation_length(B,s)
    xi_list.append(xi)


    #print "chi=%3.0f"%chi[0]," S=%.4f"%S," m=%.4f"%m,"E=%.8f"%E, "E_exact=%.8f"%E0_exact

# plotting energies, magnetization, entropy and so on
fig1 = pl.figure(figsize=[8,10])

ax1 = pl.subplot(411)
pl.plot(g_list,m_list)
pl.ylabel('$m$')

ax2 = pl.subplot(412, sharex=ax1)
pl.plot(g_list,S_list)
pl.ylabel('$S$')

ax3 = pl.subplot(413, sharex=ax1)
pl.plot(g_list,dE_list)
pl.ylabel('$\\Delta E$')

ax4 = pl.subplot(414, sharex=ax1)
pl.plot(g_list,xi_list)
pl.ylabel('$\\xi$')
pl.xlabel('$g$')

pl.show();


ii=10 # index within g_list, i.e., which coupling are we looking at
g=g_list[ii]
fig2 = pl.figure(figsize=[8,4])
pl.semilogy(ES_list[ii][1],'o')
pl.ylabel('$ES [g= %g ]$' % (g))
pl.xlabel('$\\alpha$')
pl.show();

# add the plots you prefer: 
# in particular, try to have a look at how the correlation length behaves with increasing chi :) 
# and same for the entanglement entropy...